"""Raise function names into the project namespace."""

from cluster_colors.cluster_supercluster import (
    AgglomerativeSupercluster,
    DivisiveSupercluster,
)
from cluster_colors.image_colors import get_image_clusters, show_clusters
from cluster_colors.vector_stacker import stack_vectors

__all__ = [
    "AgglomerativeSupercluster",
    "DivisiveSupercluster",
    "get_image_clusters",
    "show_clusters",
    "stack_vectors",
]



"""Try to find an address any potential non-deterministic cases irt split floats.

:author: Shay Hill
:created: 2024-09-12
"""

# pyright: reportPrivateUsage=false

import pytest

from cluster_colors.cluster_cluster import _split_floats


class TestSplitFloats:
    def test_empty(self):
        """Raise ValueError if no floats are passed."""
        with pytest.raises(ValueError):
            _ = _split_floats([])

    def test_single(self):
        """Raise ValueError if only one float is passed."""
        with pytest.raises(ValueError):
            _ = _split_floats([1.0])

    def test_identical(self):
        """Return 0 if all floats are identical."""
        assert _split_floats([1.0, 1.0, 1.0]) == 0

    def test_explicit(self):
        """Return the expected split."""
        assert _split_floats([1.0, 2.0, 3.0]) == 1

    def test_explicit_2(self):
        """Return the expected split."""
        assert _split_floats([1.0, 2.0, 2.0, 3, 3, 3, 3, 3]) == 3

    def test_explicit_3(self):
        """Return the expected split."""
        for i in range(76):
            assert _split_floats([-i, *range(100)]) == 50
        for i in range(76, 124):
            assert _split_floats([-i, *range(100)]) == 49
        for i in range(124, 169):
            assert _split_floats([-i, *range(100)]) == 48
        for i in range(169, 213):
            assert _split_floats([-i, *range(100)]) == 47
        for i in range(213, 254):
            assert _split_floats([-i, *range(100)]) == 46



"""Test module pca.py

:author: Shay Hill
:created: 2023-03-14
"""

# pyright: reportPrivateUsage=false

import math

import numpy as np

from cluster_colors import cluster_supercluster


class TestPCA:
    def test_linear(self):
        """When all points are on a line, axis points along line."""
        line = np.array([[1, 1, 1], [2, 2, 1], [3, 3, 1], [14, 14, 1]])
        cluster_ = cluster_supercluster.Cluster.from_stacked_vectors(line)
        x, y = cluster_._direction_of_highest_variance
        assert math.isclose(x, y)
        assert x != 0

    def test_square(self):
        """When all points form a square, axis points up or across."""
        square = np.array([[1, 1, 1], [2, 1, 1], [2, 2, 1], [1, 2, 1]])
        cluster_ = cluster_supercluster.Cluster.from_stacked_vectors(square)
        x, y = cluster_._direction_of_highest_variance
        assert x == 0 or y == 0
        assert x == 1 or y == 1

    def test_wide_rect(self):
        """When all points form a long rectangle, axis points across."""
        wide_rect = np.array([[1, 1, 1], [5, 1, 1], [5, 2, 1], [1, 2, 1]])
        cluster_ = cluster_supercluster.Cluster.from_stacked_vectors(wide_rect)
        np.testing.assert_allclose(cluster_._direction_of_highest_variance, (1, 0))

    def test_tall_rect(self):
        """When all points form a tall rectangle, axis points up."""
        tall_rect = np.array([[1, 1, 1], [2, 1, 1], [2, 5, 1], [1, 5, 1]])
        cluster_ = cluster_supercluster.Cluster.from_stacked_vectors(tall_rect)
        np.testing.assert_allclose(cluster_._direction_of_highest_variance, (0, 1))



"""Create and use cluster images from image colors.

:author: Shay Hill
:created: 2022-11-07
"""

# pyright: reportUnknownMemberType=false
# pyright: reportUnknownArgumentType=false

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
from PIL import Image

from cluster_colors.cluster_supercluster import DivisiveSupercluster, SuperclusterBase
from cluster_colors.config import CACHE_DIR
from cluster_colors.cut_colors import cut_colors
from cluster_colors.paths import BINARIES_DIR
from cluster_colors.pool_colors import pool_colors
from cluster_colors.vector_stacker import stack_vectors

if TYPE_CHECKING:
    from numpy import typing as npt

    from cluster_colors.type_hints import NBits, Vectors


def _stack_image_colors_no_cache(
    filename: Path | str, num_colors: int = 512, pool_bits: NBits = 6
) -> Vectors:
    """Stack pixel colors and reduce the number of colors in an image.

    :param filename: the path to an image file
     :param num_colors: the number of colors to reduce to. The default of 512 will
         cluster quickly down to medium-sized clusters.
     :param pool_bits: the number of bits to pool colors by. The default of 6 is a
     good value. You can probably just ignore this parameter, but it's here to
         eliminate a "magic number" from the code.
     :return: an array of colors with weights
    """
    img = Image.open(filename)
    img = img.convert("RGBA")
    colors = stack_vectors(np.array(img))
    colors = pool_colors(colors, pool_bits)
    return cut_colors(colors, num_colors)


def stack_image_colors(
    filename: Path | str,
    num_colors: int = 512,
    pool_bits: NBits = 6,
    *,
    ignore_cache: bool = False,
) -> Vectors:
    """Load cache or stack pixel colors and reduce the number of colors in an image.

    :param filename: the path to an image file
    :param num_colors: the number of colors to reduce to. The default of 512 will
        cluster quickly down to medium-sized clusters.
    :param pool_bits: the number of bits to pool colors by. The default of 6 is a
    good value. You can probably just ignore this parameter, but it's here to
        eliminate a "magic number" from the code.
    :param ignore_cache: if True, ignore any cached results and recompute the colors.
    :return: an array of colors with weights

    This is a pre-processing step for the color clustering. Stacking is necessary,
    and the pooling and cutting will allow clustering in a reasonable amount of time.
    """
    cache_path = CACHE_DIR / f"{Path(filename).stem}_{num_colors}_{pool_bits}.npy"
    if not ignore_cache and cache_path.exists():
        return np.load(cache_path)

    colors = _stack_image_colors_no_cache(filename, num_colors, pool_bits)
    np.save(cache_path, colors)
    return colors


def get_image_clusters(
    filename: Path | str,
    num_colors: int = 512,
    pool_bits: NBits = 6,
    *,
    ignore_cache: bool = False,
) -> DivisiveSupercluster:
    """Get all colors in an image as a single KMedSupercluster instance.

    :param filename: the path to an image file
    :param num_colors: the number of colors to reduce to. The default of 512 will
        cluster quickly down to medium-sized clusters.
    :param pool_bits: the number of bits to pool colors by. The default of 6 is a
    good value. You can probably just ignore this parameter, but it's here to
        eliminate a "magic number" from the code.
    :param ignore_cache: if True, ignore any cached results and recompute the colors.
    :return: a KMedSupercluster instance containing all the colors in the image
    """
    stacked_colors = stack_image_colors(
        filename, num_colors, pool_bits, ignore_cache=ignore_cache
    )
    return DivisiveSupercluster.from_stacked_vectors(stacked_colors)


def show_clusters(supercluster: SuperclusterBase, filename_stem: str) -> None:
    """Create a png with the exemplar of each cluster.

    :param supercluster: the clusters to show
    :param filename_stem: the filename stem to use for the output file
    """
    width = 1000
    sum_weight = sum(supercluster.members.weights)
    stripes: list[npt.NDArray[np.uint8]] = []
    for cluster in supercluster.clusters:
        stripe_width = max(round(cluster.weight / sum_weight * width), 1)
        stripes.append(
            np.tile(cluster.get_as_vector(), (800, stripe_width))
            .reshape(800, stripe_width, 3)
            .astype(np.uint8)
        )
    # combine stripes into one array
    image = np.concatenate(stripes, axis=1)

    image = Image.fromarray(image)
    image.save(BINARIES_DIR / f"{filename_stem}-{len(supercluster.clusters)}.png")



"""Show full error diffs in pytest

:author: Shay Hill
:created: 2024-09-15
"""

from typing import Any


def pytest_assertrepr_compare(config: Any, op: str, left: str, right: str):
    """See full error diffs"""
    if op in ("==", "!="):
        return [f"{left} {op} {right}"]



"""Test individual methods in the base Member, Cluster, and Supercluster classes.

:author: Shay Hill
:created: 2023-04-12
"""

import numpy as np

from cluster_colors.cluster_members import Members
from cluster_colors.cluster_supercluster import DivisiveSupercluster


class TestSupercluster:
    def test_as_one_cluster(self):
        """Test the as_one_cluster method."""
        members = Members.from_stacked_vectors(np.random.rand(6, 4))
        clusters = DivisiveSupercluster(members)
        assert len(clusters.get_as_stacked_vectors()) == 1



"""Type hints for cluster_colors.

:author: Shay Hill
:created: 2022-10-22
"""

from collections.abc import Sequence
from typing import Annotated, Any, Literal, TypeAlias

import numpy as np
from numpy import typing as npt

# input pixel array, or something that looks like an input pixel array.
Pixels: TypeAlias = Annotated[npt.NDArray[np.number[Any]], "(..., -1)"]


# array that has been cast to float, but it not expected to have a weight axis or
# particular shape.
FPArray: TypeAlias = npt.NDArray[np.float64]

# a 1D array of floats
Vector: TypeAlias = Annotated[FPArray, (-1,)]
# something that can be cast to a vector
VectorLike: TypeAlias = Sequence[float] | Vector

# an array of vectors, (x, ...) or (x, ..., w)
Vectors: TypeAlias = Annotated[npt.NDArray[np.float64], (-1, -1)]
# something that can be cast to vectors
VectorsLike: TypeAlias = Sequence[Sequence[float]] | Vectors


ProximityMatrix: TypeAlias = Annotated[FPArray, "(n, n)"]

# Names of available centers
CenterName = Literal["weighted_medoid", "unweighted_medoid", "weighted_median"]


QualityMetric = Literal["sum_error", "max_error", "avg_error", "span"]
CentroidName = Literal["weighted_medoid", "unweighted_medoid"]


# number of bits in a color channel
NBits = Literal[1, 2, 3, 4, 5, 6, 7, 8]



"""Test methods in KMedDivisiveSupercluster

:author: Shay Hill
:created: 2023-03-14
"""

# pyright: reportPrivateUsage=false
# pyright: reportUnknownMemberType=false
# pyright: reportUnknownVariableType=false
# pyright: reportMissingParameterType=false
# pyright: reportUnknownParameterType=false

from typing import Annotated

import numpy as np
import numpy.typing as npt
import pytest

from cluster_colors.cluster_supercluster import (
    AgglomerativeSupercluster,
    DivisiveSupercluster,
)
from cluster_colors.vector_stacker import stack_vectors

ColorsArray = Annotated[npt.NDArray[np.float64], (-1, 3)]


@pytest.fixture(
    scope="function",
    params=[np.random.randint(0, 255, (100, 4), dtype=np.uint8) for _ in range(10)],
)
def colors(request: pytest.FixtureRequest) -> ColorsArray:
    return stack_vectors(request.param)


class TestKMedians:
    def test_get_rsorted_clusters(self, colors: ColorsArray):
        """Test that the clusters are sorted by the number of colors in them"""
        clusters = DivisiveSupercluster.from_stacked_vectors(colors)
        clusters.set_n(16)
        _ = clusters.get_as_stacked_vectors()

    def test_split_to_n(self, colors: ColorsArray):
        clusters = DivisiveSupercluster.from_stacked_vectors(colors)
        clusters.set_n(10)
        assert clusters.get_as_stacked_vectors().shape == (10, 4)

    def test_get_rsorted_exemplars(self, colors: ColorsArray):
        """Test that the clusters are sorted by the number of colors in them"""
        clusters = DivisiveSupercluster.from_stacked_vectors(colors)
        clusters.set_n(16)
        assert clusters.get_as_vectors().shape == (16, 3)

    def test_merge_to_n(self, colors: ColorsArray):
        clusters = AgglomerativeSupercluster.from_stacked_vectors(colors[:24])
        clusters.set_n(10)
        assert clusters.get_as_stacked_vectors().shape == (10, 4)

    def test_from_cluster_subset(self, colors: ColorsArray):
        clusters = DivisiveSupercluster.from_vectors(colors)
        clusters.set_n(8)
        subset = DivisiveSupercluster.from_cluster_subset(*clusters.clusters[:4])
        subset_cnt = sum(len(c.ixs) for c in clusters.clusters[:4])
        assert subset.members.vectors.shape == (subset_cnt, 3)
        assert subset.members.weights.shape == (subset_cnt,)
        assert subset.members.pmatrix.shape == (subset_cnt, subset_cnt)


class TestPredicates:
    def test_set_max_sum_error(self, colors: ColorsArray):
        """Split as far as necessary to get below the threshold"""
        clusters = DivisiveSupercluster.from_stacked_vectors(colors)
        for _ in range(10):
            clusters.split()
        max_sum_error = 48000
        clusters.set_max_sum_error(max_sum_error)
        assert clusters.get_max_sum_error() <= max_sum_error
        clusters.merge()
        assert clusters.get_max_sum_error() > max_sum_error

    def test_set_max_span(self, colors: ColorsArray):
        """Split as far as necessary to get below the threshold"""
        clusters = DivisiveSupercluster.from_stacked_vectors(colors)
        for _ in range(10):
            clusters.split()
        max_span = 2400
        clusters.set_max_span(max_span)
        assert clusters.get_max_span() <= max_span
        clusters.merge()
        assert clusters.get_max_span() > max_span

    def test_set_max_max_error(self, colors: ColorsArray):
        """Split as far as necessary to get below the threshold"""
        clusters = DivisiveSupercluster.from_stacked_vectors(colors)
        for _ in range(10):
            clusters.split()
        max_max_error = 1200
        clusters.set_max_max_error(max_max_error)
        assert clusters.get_max_max_error() <= max_max_error
        clusters.merge()
        assert clusters.get_max_max_error() > max_max_error

    def test_set_avg_error(self, colors: ColorsArray):
        """Split as far as necessary to get below the threshold"""
        clusters = DivisiveSupercluster.from_stacked_vectors(colors)
        for _ in range(10):
            clusters.split()
        max_avg_error = 0.5
        clusters.set_max_avg_error(max_avg_error)
        assert clusters.get_max_avg_error() <= max_avg_error
        clusters.merge()
        assert clusters.get_max_avg_error() > max_avg_error



"""Test functions in triangulate_image.kmedians.py

:author: Shay Hill
:created: 2022-09-16
"""

import numpy as np
from matplotlib import pyplot as plt

# pyright: reportPrivateUsage=false
from cluster_colors.cluster_members import Members
from cluster_colors.cluster_supercluster import (
    AgglomerativeSupercluster,
    Cluster,
    DivisiveSupercluster,
    SuperclusterBase,
)


class TestClusterExemplar:
    """Test triangulate_image._Cluster.exemplar property"""

    def test_weighted_medoid(self) -> None:
        """Return weighted average of member.rgb values."""
        vectors = np.array([[1, 2, 3, 1], [4, 5, 6, 2], [7, 8, 9, 3]])
        members = Members.from_stacked_vectors(vectors)
        cluster = Cluster(members, [0, 2])
        np.testing.assert_array_equal(cluster.weighted_medoid, 2)

    def test_unweighted_medoid(self) -> None:
        """Return the member with lowest cost, ignoring weights."""
        vectors = np.array([[1, 2, 3, 1], [4, 5, 6, 2], [7, 8, 9, 3]])
        members = Members.from_stacked_vectors(vectors)
        cluster = Cluster(members, [0, 1, 2])
        np.testing.assert_array_equal(cluster.unweighted_medoid, 1)

    def test_weighted_median(self) -> None:
        """Return the weighted median of the members."""
        vectors = np.array([[1, 2, 3, 1], [4, 5, 6, 2], [7, 8, 9, 3]])
        members = Members.from_stacked_vectors(vectors)
        cluster = Cluster(members, [0, 1, 2])
        np.testing.assert_array_equal(cluster.weighted_median, (5.5, 6.5, 7.5))


class TestCluster:

    def test_split(self) -> None:
        """Return 256 clusters given 256 colors."""
        vectors = np.random.rand(50, 4) * 255
        members = Members.from_stacked_vectors(vectors)
        cluster = Cluster(members, range(50))
        child_a, child_b = cluster.split()
        assert set(child_a.ixs) | set(child_b.ixs) == set(cluster.ixs)

    def test_plot_2d_clusters(self) -> None:
        """Display clusters as a scatter plot."""
        vectors = np.random.rand(150, 3) * 255
        supercluster = DivisiveSupercluster.from_stacked_vectors(vectors)
        for _ in range(4):
            supercluster.split()
        plot_2d_clusters(supercluster)

    def test_plot_2d_clusters_agglomerative(self) -> None:
        """Display clusters as a scatter plot.

        This test occasionally fails with a _tkinter.TclError. I'm not sure why, but
        re-running with the same output works, so it's not something I'm going to try
        to fix. So, I catch the error, print it, and move on.
        """
        vectors = np.random.rand(150, 3) * 255
        supercluster = AgglomerativeSupercluster.from_stacked_vectors(vectors)
        supercluster.set_n(5)

        try:
            plot_2d_clusters(supercluster)
        except Exception as e:
            print(e)


def plot_2d_clusters(supercluster: SuperclusterBase) -> None:
    """Display clusters as a scatter plot.

    :param supercluster: list of sets of (x, y) coordinates

    Make each cluster a different color.
    """
    clusters = supercluster.clusters
    # colors = plt.cm.rainbow(np.linspace(0, 1, len(clusters)))
    # colors2 = stack_vectors(colors)

    clusters = sorted(clusters, key=lambda c: c.weight)[::-1]
    colors = np.zeros((len(clusters), 4))
    colors[:, 3] = 1
    colors[:, :2] = supercluster.get_as_vectors() / 255
    colors[:, 2] = 1 - np.max(colors[:, :2], axis=1)

    for cluster, color in zip(clusters, colors):
        points = cluster.members.vectors[cluster.ixs]
        exemplar = cluster.get_as_vector("unweighted_medoid")
        plt.scatter([exemplar[0]], [exemplar[1]], color="black", s=100)  # type: ignore
        xs = [x for x, _ in points]
        ys = [y for _, y in points]
        plt.scatter(xs, ys, color=color)  # type: ignore
    plt.show()  # type: ignore



"""Cluster colors with median cut.

Repeatedly subdivide the color space by splitting along the longest axis (not
constrained to x, y, or z. splits with the optimal plane).

This median cut has a few optimizations, so it will produce a nice heirarchal set of
clusters with some modification, though at this point no hierarchies are preserved.
The idea is to use this to reduce the number of colors in an image to around 512.
From 512, they can be again merged into a single cluster and again split to find a
starting point for kmedians.

:author: Shay Hill
:created: 2022-10-21
"""

import numpy as np

from cluster_colors.cluster_cluster import Cluster
from cluster_colors.type_hints import Vectors


def _split_every_cluster(clusters: set[Cluster]) -> set[Cluster]:
    """Recursively split every cluster.

    :param clusters: A set of clusters.
    :returns: A set of clusters.

    Recursively split every cluster with no regard for error. Will only *not* split
    a cluster if it only has one member.
    """
    splittable = {c for c in clusters if len(c.ixs) > 1}
    if not splittable:
        return clusters
    for cluster in splittable:
        clusters.remove(cluster)
        clusters.update(cluster.split())
    return clusters


def _split_largest_cluster(clusters: set[Cluster], num: int) -> set[Cluster]:
    """Split one cluster per call.

    :param clusters: A set of clusters.
    :returns: A set of clusters.
    """
    if len(clusters) >= num:
        return clusters

    next_split = max(clusters, key=lambda c: c.span)
    if next_split.span == 0:
        return clusters
    clusters.remove(next_split)
    clusters.update(next_split.split())
    return clusters


def cut_colors(colors: Vectors, num: int) -> Vectors:
    """Merge colors into a set of num colors.

    :param colors: a (-1, 4) array of unique rgb values with weights
    :param num: the number of colors to split into
    :returns: a (-1, 4) array of unique rgb values with weights

    Put all colors into one cluster, split that cluster into num clusters, then
    return a median color for each cluster.

    Splits every cluster until roughly half the requested number of clusters have
    been created, then starts cherry picking. This idea was proposed and tested in a
    paper I ran into online, but I can't find it now.
    """
    if len(colors) <= num:
        return colors

    clusters = {Cluster.from_stacked_vectors(colors)}
    while len(clusters) < num // 4:
        clusters = _split_every_cluster(clusters)

    while len(clusters) < num:
        len_clusters = len(clusters)
        clusters = _split_largest_cluster(clusters, num)
        if len(clusters) == len_clusters:
            break

    return np.array([c.get_as_stacked_vector() for c in clusters])



"""Cache directory for the project.

:author: Shay Hill
:created: 2023-03-13
"""

from pathlib import Path
from tempfile import TemporaryFile

with TemporaryFile() as f:
    CACHE_DIR = Path(f.name).parent / "cluster_colors_cache"

CACHE_DIR.mkdir(parents=True, exist_ok=True)



"""Test reduce_colors() function.

:author: Shay Hill
:created: 2022-09-19
"""

# pyright: reportPrivateUsage=false
# pyright: reportUnknownMemberType=false
# pyright: reportUnknownVariableType=false
# pyright: reportUnknownParameterType=false
# pyright: reportUnknownArgumentType=false
# pyright: reportMissingParameterType=false

import random
from typing import TypeVar

import numpy as np
from PIL import Image

from cluster_colors import pool_colors
from cluster_colors.cluster_supercluster import DivisiveSupercluster
from cluster_colors.paths import TEST_DIR
from cluster_colors.vector_stacker import stack_vectors

_T = TypeVar("_T")

DIAG = np.array([(i, i, i, i + 1) for i in range(256)])

img = Image.open(TEST_DIR / "sugar-shack-barnes.jpg")
colors = np.array(img).reshape(-1, 1)[::16]
stacked_colors = stack_vectors(colors)


class TestPoolColors:
    def test_sum_weight(self):
        """Total weight of all colors is number of pixels."""
        img = Image.open(TEST_DIR / "sugar-shack-barnes.jpg")
        colors = np.array(img)[:100, :100]
        weights = np.full(colors.shape[:-1], 1, dtype=float)
        weighted_colors = stack_vectors(np.dstack((colors, weights)))
        reduced = pool_colors.pool_colors(weighted_colors, 4)
        assert np.sum(reduced[..., 3]) == colors.shape[0] * colors.shape[1]

    def test_robust_to_order(self):
        """Order of colors should not matter."""
        reduced = {tuple(x) for x in pool_colors.pool_colors(stacked_colors, 4)}
        reduced2 = {tuple(x) for x in pool_colors.pool_colors(stacked_colors[::-1], 4)}
        assert reduced == reduced2

    def test_robust_to_order_with_ties(self):
        """Order does not matter, even with ties."""

        def split_ten_times(vecs):
            clusters = DivisiveSupercluster.from_vectors(np.array(vecs))
            for _ in range(10):
                clusters.split()
            return {tuple(sorted(x.ixs)) for x in clusters.clusters}

        vecs_ = [(x, y, z) for x in range(5) for y in range(4) for z in range(3)]
        expect = split_ten_times(vecs_)
        for _ in range(10):
            random.shuffle(vecs_)
            result = split_ten_times(vecs_)
            assert result == expect

    def test_singles(self):
        """When color has a weight of 1 and does not stack, return same."""
        reduced = {tuple(x) for x in pool_colors.pool_colors(colors, 4)}
        reduced2 = {tuple(x) for x in pool_colors.pool_colors(colors[::-1], 4)}
        assert reduced == reduced2

    def test_small_number_of_identical_colors(self):
        """Do not return identical colors.

        Previousy, `pool_colors` would not try to pool when there were fewer input
        colors than the number of boxes in an nbits cube. This is to avoid needlessly
        discarding colors when few colors are present. This behavior caused problems:

        * identical colors were stacked by `stack_vectors`
        * *nearly* identical colors were not stacked or pooled.

        This caused problems later on, because the nearly identical colors *were*
        identical at 8-bits. Eventually, these colors created identical clusters,
        which broke the clustering algorithm.
        """
        colors = np.array(
            [
                [254.8, 0, 0, 1],
                [0, 254.9, 0, 1],
                [0, 254.8, 0, 1],
                [0, 254.7, 0, 1],
                [0, 0, 255, 1],
                [0, 0, 254.9, 1],
                [0, 0, 254.8, 1],
                [0, 0, 254.7, 1],
                [0, 0, 254.6, 1],
            ]
        )
        stacked_colors = stack_vectors(colors)
        reduced = sorted(map(tuple, pool_colors.pool_colors(stacked_colors, 4)))
        expect = [[0, 0, 254.8, 5], [0, 254.8, 0, 3], [254.8, 0, 0, 1]]
        np.testing.assert_array_almost_equal(reduced, expect)



"""Exceptions where clusters can no longer be split or merged.

:author: Shay Hill
:created: 2024-09-16
"""


class FailedToSplitError(Exception):
    """Exception raised when no clusters can be split."""

    def __init__(self, message: str | None = None) -> None:
        """Create a new AllClustersAreSingletonsError instance."""
        message_ = message or "Cannot split any cluster. All clusters are singletons."
        self.message = message_
        super().__init__(self.message)


class FailedToMergeError(Exception):
    """Exception raised when no clusters can be merged."""

    def __init__(self, message: str | None = None) -> None:
        """Create a new CannotMergeSingleCluster instance."""
        message_ = message or "Cannot merge any cluster. All members in one cluster."
        self.message = message_
        super().__init__(self.message)



"""Members of a cluster represented as a list of vectors with a proximity matrix.

The clustering algorithms in this package do not calculate proximity, but only refer
to a given proximity matrix. This means that more expensive proximity calculations
can be used with the minimum penalty. Each Supercluster and the Clusters it contains
share a single Members instance. This instance holds all vector information and a few
variants of the proximity matrix. A Cluster is defined by the indices of the members
it contains.

:author: Shay Hill
:created: 2024-09-01
"""

from __future__ import annotations

import functools
from typing import TYPE_CHECKING

import numpy as np
from basic_colormath import get_sqeuclidean_matrix
from paragraphs import par

from cluster_colors.vector_stacker import add_weight_axis, stack_vectors

if TYPE_CHECKING:
    from collections.abc import Iterable

    from cluster_colors.type_hints import FPArray, ProximityMatrix, VectorsLike


class Members:
    """A list of cluster vectors with a proximity matrix."""

    def __init__(
        self,
        vectors: VectorsLike,
        *,
        weights: Iterable[float] | float | None = None,
        pmatrix: ProximityMatrix | None = None,
    ) -> None:
        """Create a new Members instance.

        :param vectors: array (n, m) of vectors
        :param weights: optional array (n,) of weights
        :param pmatrix: optional proximity matrix. If not given, will be calculated
            with squared Euclidean distance
        :raises ValueError: if input vectors and pmatrix do not align after stacking

        The `weights` argument is optional, but each member must have a weight. If
        `weights` is None, the weights are assumed to be the last axis of the vectors
        argument. This is consistent with `stacked_vectors` or image colors as
        `(r,g,b,a)`. If weights is a scalar, all members are given that weight. This
        is useful for clustering non-color vectors or other instances where all
        vectors should be treated with the same priority.

        The vectors must be sorted in case any members has equal proximity to
        multiple centers. When this happens, the centroid with the lowest
        lexigraphical order is chosen.

        Init will sort input vectors and sort any proximity matrix passed to align to
        the sorted vectors.

        Init will also stack vectors to remove any duplicates, but *cannot* update an
        input pmatrix to match. If you pass identical vectors and a pmatrix, you will
        get a ValueError.
        """
        self._stacked_vectors = np.asarray(vectors)

        if weights is None:
            pass
        elif isinstance(weights, int | float):
            self._stacked_vectors = add_weight_axis(self._stacked_vectors, weights)
        else:
            weights = np.asarray(list(weights)).reshape(-1, 1)
            self._stacked_vectors = np.hstack((self._stacked_vectors, weights))

        # sort input vectors
        sort_indices = np.lexsort(self._stacked_vectors.T[::-1])
        self._stacked_vectors = self._stacked_vectors[sort_indices]

        # stack input vectors to remove duplicates
        ensure_stacked = stack_vectors(self._stacked_vectors)
        if self._stacked_vectors.shape != ensure_stacked.shape and pmatrix is not None:
            msg = par(
                """Input pmatrix shape does not conform to input vectors after
                stacking duplicates."""
            )
            raise ValueError(msg)
        self._stacked_vectors = ensure_stacked

        if pmatrix is None:
            self.pmatrix = get_sqeuclidean_matrix(self.vectors)
        else:
            self.pmatrix = pmatrix[np.ix_(sort_indices, sort_indices)]

    def __len__(self) -> int:
        """Number of members in the Members instance.

        :return: number of members
        """
        return len(self.vectors)

    @property
    def vectors(self) -> FPArray:
        """Array of vectors.

        :return: array of vectors
        """
        return self._stacked_vectors[:, :-1]

    @property
    def weights(self) -> FPArray:
        """Array of weights.

        :return: array of weights
        """
        return self._stacked_vectors[:, -1]

    @functools.cached_property
    def weighted_pmatrix(self) -> ProximityMatrix:
        """Proximity matrix with weights applied.

        :return: proximity matrix such that sum(pmatrix[i, (j, k, ...)]) is the cost
            of members[i] in a cluster with members[i, j, k, ...]
        """
        weight_columns = np.tile(self.weights, (len(self.weights), 1))
        return self.pmatrix * weight_columns

    @functools.cached_property
    def pmatrix_with_inf_diagonal(self) -> ProximityMatrix:
        """Proximity matrix with infinity on the diagonal.

        :return: proximity matrix with infinity on the diagonal. The is useful for
            finding the minumum proximity between members that is *not* the distance
            between a member and itself.
        """
        pmatrix_copy = self.pmatrix.copy()
        np.fill_diagonal(pmatrix_copy, np.inf)
        return pmatrix_copy

    # ===========================================================================
    #   constructors
    # ===========================================================================

    @classmethod
    def from_vectors(
        cls, vectors: VectorsLike, *, pmatrix: FPArray | None = None
    ) -> Members:
        """Create a Members instance from stacked_vectors.

        :param vectors: (n, m) a list of vectors
        :param pmatrix: optional proximity matrix. If not given, will be calculated
            with squared Euclidean distance
        :return: Members instance
        """
        return cls(vectors, weights=1, pmatrix=pmatrix)

    @classmethod
    def from_stacked_vectors(
        cls, stacked_vectors: VectorsLike, *, pmatrix: FPArray | None = None
    ) -> Members:
        """Create a Members instance from stacked_vectors.

        :param stacked_vectors: (n, m + 1) a list of vectors with weight channels in
            the last axis
        :param pmatrix: optional proximity matrix. If not given, will be calculated
            with squared Euclidean distance
        :return: Members instance
        """
        return cls(stacked_vectors, weights=None, pmatrix=pmatrix)



"""Test image_colors module."""

from cluster_colors import image_colors
from cluster_colors.cluster_supercluster import DivisiveSupercluster
from cluster_colors.paths import TEST_DIR

_TEST_IMAGE = TEST_DIR / "sugar-shack-barnes.jpg"


class TestGetBiggestColor:

    def test_display(self):
        """Test display_biggest_color function."""
        colors = image_colors.stack_image_colors(_TEST_IMAGE)
        clusters = DivisiveSupercluster.from_stacked_vectors(colors)
        clusters.set_n(2)
        clusters.set_max_max_error(64**2)
        image_colors.show_clusters(clusters, _TEST_IMAGE.stem)



"""Test the vector_stacker module.

:author: Shay Hill
:created: 2022-10-18
"""

# pyright: reportPrivateUsage=false
# pyright: reportUnknownMemberType=false
# pyright: reportUnknownVariableType=false
# pyright: reportMissingParameterType=false
# pyright: reportUnknownParameterType=false

import numpy as np

from cluster_colors import vector_stacker as sv


class TestAddWeightAxis:
    """Test the add_weight_axis method."""

    def test_add_weight_axis(self):
        """Test the add_weight_axis method."""
        np.testing.assert_array_equal(
            sv.add_weight_axis(np.array([[1, 2, 3], [4, 5, 6]])),
            np.array([[1, 2, 3, 255], [4, 5, 6, 255]]),
        )

    def test_optional_weight_parameter(self):
        """Test the optional weight parameter."""
        np.testing.assert_array_equal(
            sv.add_weight_axis(np.array([[1, 2, 3], [4, 5, 6]]), 128),
            np.array([[1, 2, 3, 128], [4, 5, 6, 128]]),
        )

    def test_add_weight_axis_does_not_alter_input(self):
        """Test the add_weight_axis method does not alter input."""
        input_array = np.array([[1, 2, 3], [4, 5, 6]])
        _ = sv.add_weight_axis(input_array)
        np.testing.assert_array_equal(input_array, np.array([[1, 2, 3], [4, 5, 6]]))

    def test_uint8_input_returns_floats_in_output(self):
        """Test the add_weight_axis method returns floats for uint8 input."""
        output = sv.add_weight_axis(
            np.array([[1, 2, 3], [4, 5, 6]], dtype=np.uint8), weight=np.uint8(128)  # type: ignore
        )
        assert output.dtype == np.float64


class TestStackVectors:
    def test_stack_unique_vectors(self):
        """Test the stack_unique_vectors method."""
        np.testing.assert_array_equal(
            sv.stack_vectors(np.array([[1, 2, 3, 55], [4, 5, 6, 5], [1, 2, 3, 45]])),
            np.array([[1, 2, 3, 100], [4, 5, 6, 5]]),
        )

    def test_uint8_input_returns_floats_in_output(self):
        """Test the stack_unique_vectors method returns floats for uint8 input."""
        output = sv.stack_vectors(
            np.array([[1, 2, 3, 55], [4, 5, 6, 5], [1, 2, 3, 45]], dtype=np.uint8)
        )
        assert output.dtype == np.float64

    def test_stack_vectors_does_not_alter_input(self):
        """Test the stack_unique_vectors method does not alter input."""
        input_array = np.array([[1, 2, 3, 55], [4, 5, 6, 5], [1, 2, 3, 45]])
        _ = sv.stack_vectors(input_array)
        np.testing.assert_array_equal(
            input_array, np.array([[1, 2, 3, 55], [4, 5, 6, 5], [1, 2, 3, 45]])
        )

    def test_optional_weight_parameter(self):
        """Test the optional weight parameter."""
        np.testing.assert_array_equal(
            sv.stack_vectors(np.array([[1, 2, 3], [4, 5, 6], [1, 2, 3]]), weight=128),
            np.array([[1, 2, 3, 256], [4, 5, 6, 128]]),
        )


class TestStackColors:
    def test_stack_vectors(self):
        """Test the stack_vectors method on 4-axis vectors."""
        np.testing.assert_array_equal(
            sv.stack_vectors(np.array([[1, 2, 3, 55], [4, 5, 6, 5], [1, 2, 3, 45]])),
            np.array([[1, 2, 3, 100], [4, 5, 6, 5]]),
        )

    def test_stack_single_vectors(self):
        """Test the stack_vectors method on 1-axis vectors."""
        np.testing.assert_array_equal(
            sv.stack_vectors(np.array([[1, 1], [4, 1], [1, 1]])),
            np.array([[1, 2], [4, 1]]),
        )



"""Type for a single cluster.

Each Cluster instance holds a Members instance and a list of indices to vectors in
that instance. Clusters have multiple errors and centers defined for various
purposes, including two centroids, weighted and unweighted medoid, used within the
clustering algorithm.

:author: Shay Hill
:created: 2024-09-03
"""

from __future__ import annotations

import functools
import itertools
from typing import TYPE_CHECKING

import numpy as np
from paragraphs import par
from stacked_quantile import get_stacked_medians

from cluster_colors.cluster_members import Members

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable

    from cluster_colors.type_hints import (
        CenterName,
        CentroidName,
        FPArray,
        ProximityMatrix,
        QualityMetric,
        Vector,
        VectorsLike,
    )


def _construct_is_low(low: float, high: float) -> Callable[[float], bool]:
    """Return a function that determines if a value is closer to below than above.

    :param low: exemplar of low values
    :param hight: exemplar of high values
    :return: function that determines if a value is closer to low than high
    """

    def is_low(value: float) -> bool:
        """Determine if a value is closer to low than high.

        :param value: float
        :return: True if value is closer to low than high
        """
        return abs(value - low) < abs(value - high)

    return is_low


def _split_floats(floats: Iterable[float]) -> int:
    """Find an index of a sorted list of floats that minimizes the sum of errors.

    :param floats: An iterable of float numbers.
    :return: An index of the list that minimizes the sum of errors.

    This is used to find a natural split along the axis of highest variance, if such
    exists. During the k-medoids clustering, this little bit of intelligence will
    matter less, because the SuperclusterBase._reassign method will help address poor
    splits, but a few hundredths of a second is worth the effort during the median
    cut (`cut_colors`) process.
    """
    floats = sorted(floats)
    if len(floats) < 2:
        msg = "Cannot split a list of floats with fewer than 2 elements"
        raise ValueError(msg)

    def converge(splitter: int) -> int:
        if splitter == 0:  # all floats are identical
            return 0
        below, above = floats[:splitter], floats[splitter:]
        is_low = _construct_is_low(sum(below) / len(below), sum(above) / len(above))
        new_splitter = len(list(itertools.takewhile(is_low, floats)))
        if new_splitter == splitter:
            return splitter
        return converge(new_splitter)

    return converge(len(floats) // 2)


class Cluster:
    """A cluster of indices to self.members.vectors.

    quality_metric: The property used to select which cluster to split. This is
        meaningles for agglomerative clustering.
    quality_centroid: The property used as a centroid when computing the quality
        metric.
    """

    quality_metric: QualityMetric = "sum_error"
    quality_centroid: CentroidName = "weighted_medoid"

    def __init__(self, members: Members, ixs: Iterable[int] | None = None) -> None:
        """Identify a cluster by the indices of its members.

        :param members: Members instance
        :param ixs: optional indices of members. If None, use all members.
        """
        self.members = members
        if ixs is None:
            self.ixs = np.arange(len(self.members), dtype=np.int32)
        else:
            self.ixs = np.array(sorted(ixs), dtype=np.int32)

    # ===========================================================================
    #   parameter-based properties
    # ===========================================================================

    @property
    def error(self) -> float:
        """Get the value of the split metric.

        :return: value of the split metric
        """
        return getattr(self, self.quality_metric)

    @property
    def centroid(self) -> int:
        """Get the value of the error centroid.

        :return: value of the error centroid
        """
        return getattr(self, self.quality_centroid)

    # ===========================================================================
    #   constructors
    # ===========================================================================

    @classmethod
    def from_vectors(
        cls, vectors: VectorsLike, pmatrix: ProximityMatrix | None = None
    ) -> Cluster:
        """Create a Cluster instance from a sequence or array of colors.

        :param vectors: An iterable of vectors
            [(r0, g0, b0), (r1, g1, b1), ...]
        :param pmatrix: optional proximity matrix. If not given, will be calculated
            with squared Euclidean distance
        :return: A Cluster instance with members
            {Member([r0, g0, b0, 1]), Member([r1, g1, b1, 1]), ...}
        """
        members = Members.from_vectors(vectors, pmatrix=pmatrix)
        return cls(members)

    @classmethod
    def from_stacked_vectors(
        cls, stacked_vectors: VectorsLike, pmatrix: ProximityMatrix | None = None
    ) -> Cluster:
        """Create a Cluster instance from an array of colors with a weight axis.

        :param stacked_vectors: An iterable of vectors with a weight axis
            [(r0, g0, b0, w0), (r1, g1, b1, w1), ...]
        :param pmatrix: optional proximity matrix. If not given, will be calculated
            with squared Euclidean distance
            [(r0, g0, b0, w0), (r1, g1, b1, w1), ...]
        :return: A Cluster instance with members
            {Member([r0, g0, b0, w0]), Member([r1, g1, b1, w1]), ...}
        """
        members = Members.from_stacked_vectors(stacked_vectors, pmatrix=pmatrix)
        return cls(members)

    # ===========================================================================
    #   vector-like properties
    # ===========================================================================

    @property
    def weight(self) -> float:
        """Total weight of members.

        :return: total weight of members
        """
        return sum(self.members.weights[self.ixs])

    @property
    def error_weight(self) -> float:
        """Total weight of members that are not the center.

        :return: total weight of members that are not the center

        This is for use in various metrics one might use for selecting clusters to
        split or join.
        """
        if len(self.ixs) == 1:
            return 0
        return self.weight - self.members.weights[self.centroid]

    def get_as_vector(self, which_center: CenterName | None = None) -> Vector:
        """Get the exemplar as a vector.

        :param which_center: optionally specify a cluster center attribute. Choices
            are 'weighted_median', 'weighted_medoid', or 'unweighted_medoid'. Default
            is 'weighted_median'.
        :return: cluster center as a vector
        """
        which_center = which_center or "weighted_median"
        if which_center == "weighted_median":
            return self.weighted_median
        if which_center == "weighted_medoid":
            return self.members.vectors[self.weighted_medoid]
        if which_center == "unweighted_medoid":
            return self.members.vectors[self.unweighted_medoid]
        msg = par(
            f"""center_name must be 'weighted_median', 'weighted_medoid', or
            'unweighted_medoid', but got {which_center}"""
        )
        raise ValueError(msg)

    def get_as_stacked_vector(self, which_center: CenterName | None = None) -> Vector:
        """Get the exemplar as a stacked vector.

        :return: [*center, weight] as a stacked vector
        :param which_center: optionally specify a cluster center attribute. Choices
            are 'weighted_median', 'weighted_medoid', or 'unweighted_medoid'. Default
            is 'weighted_median'.
        """
        vector = self.get_as_vector(which_center)
        weight = sum(self.members.weights[self.ixs])
        return np.append(vector, weight)

    # ===========================================================================
    #   cluster centers
    # ===========================================================================

    def _get_weighted_medoid(self, ixs: Iterable[int] | None = None) -> int:
        """Get the index of the mediod, respecting weights.

        :param ixs: optional subset of members indices. I can't see a use case for
            manually passing this, but it's here to break ties in property
            unweighted_medoid.
        :return: index of the mediod, respecting weights
        """
        ixs_ = self.ixs if ixs is None else np.array(list(ixs), dtype=np.int32)
        if len(ixs_) == 1:
            return int(ixs_[0])
        pmatrix = self.members.weighted_pmatrix[np.ix_(ixs_, ixs_)]
        return ixs_[np.argmin(pmatrix.sum(axis=1))]

        return int(ixs_[np.argmin(self.members.weighted_pmatrix[ixs_].sum(axis=1))])

    @functools.cached_property
    def weighted_medoid(self) -> int:
        """Get cluster exemplar.

        :return: the index of the exemplar with the least cost

        If I strictly followed my own conventions, I'd just call this property `vs`,
        but this value acts as the exemplar when clustering, so I prefer to use this
        alias in my clustering code.
        """
        return self._get_weighted_medoid()

    @functools.cached_property
    def unweighted_medoid(self) -> int:
        """Get the index of the mediod, mostly ignoring weights.

        :return: index of the mediod

        If multiple members are tied for cost, use weights to break the tie. This
        will always be the case with two members, but it is theoretically possible
        with more members. That won't happen, but it's cheap to cover the case.
        """
        if len(self.ixs) < 3:
            return self._get_weighted_medoid()

        row_sums = self.members.pmatrix[np.ix_(self.ixs, self.ixs)].sum(axis=1)
        min_cost = np.min(row_sums)
        arg_where_min = np.argwhere(row_sums == min_cost).flatten()
        arg_where_min = [self.ixs[x] for x in arg_where_min]

        if len(arg_where_min) == 1:
            return int(arg_where_min[0])
        return self._get_weighted_medoid(map(int, arg_where_min))

    @functools.cached_property
    def weighted_median(self) -> Vector:
        """Get the median of the cluster, respecting weights.

        :return: median of the cluster

        This is categorically different than the medoid, because the median is not a
        member of the cluster. So this property is not an index to the cluster
        members, but a vector that is likely not coincident with any member.
        """
        weights = self.members.weights[self.ixs].reshape(-1, 1)
        return get_stacked_medians(self.members.vectors[self.ixs], weights)

    # ===========================================================================
    #   covariance matrix and eigenvectors
    # ===========================================================================

    @property
    def _covariance_matrix(self) -> FPArray:
        """Get the covariance matrix of the cluster.

        :return: covariance matrix of the cluster
        """
        vs = self.members.vectors
        ws = np.ceil(self.members.weights)
        return np.cov(vs.T, fweights=ws)

    @functools.cached_property
    def _np_linalg_eig(self) -> tuple[FPArray, FPArray]:
        """Cache the value of np.linalg.eig on the covariance matrix of the cluster.

        :return: tuple of eigenvalues and eigenvectors
        """
        eigenvalues, eigenvectors = np.linalg.eig(self._covariance_matrix)
        return np.real(eigenvalues), np.real(eigenvectors)

    @property
    def _direction_of_highest_variance(self) -> FPArray:
        """Get the first Eigenvector of the covariance matrix.

        :return: first Eigenvector of the covariance matrix

        Return the normalized eigenvector with the largest eigenvalue.
        """
        eigenvalues, eigenvectors = self._np_linalg_eig
        return eigenvectors[:, np.argmax(eigenvalues)]

    # ===========================================================================
    #   quality metrics
    # ===========================================================================

    @functools.cached_property
    def sum_error(self) -> float:
        """Get the sum of proximity errors of all members.

        :return: sum of errors of all members
        """
        if len(self.ixs) == 1:
            return 0
        pmatrix = self.members.weighted_pmatrix
        weights = pmatrix[self.centroid, self.ixs]
        return float(np.sum(weights))

    @functools.cached_property
    def max_error(self) -> float:
        """Get the max of proximity errors of all members.

        :return: max of proximity errors of all members

        As elsewhere, weights are treated as frequencies, so the maximum error is the
        error of the most distant member, regardless of weight, even if that weight
        is 0.
        """
        if len(self.ixs) == 1:
            return 0
        pmatrix = self.members.pmatrix
        weights = pmatrix[self.centroid, self.ixs]
        return float(np.max(weights))

    @functools.cached_property
    def span(self) -> float:
        """Get the maximum proximity error between any two members.

        :return: maximum squared error of all members
        """
        if len(self.ixs) == 1:
            return 0
        pmatrix = self.members.pmatrix
        weights = pmatrix[np.ix_(self.ixs, self.ixs)]
        return float(np.max(weights))

    @functools.cached_property
    def avg_error(self) -> float:
        """Get the average error of the cluster members.

        :return: max_error / error_weight

        This metric is useful for splitting clusters with meaningful outliers, where
        sum_error might split heavy clusters, even when the members are close.
        """
        if self.error_weight == 0:
            return 0
        return self.sum_error / self.error_weight

    # ===========================================================================
    #   examine merge candidates
    # ===========================================================================

    def get_merge_span(self, other: Cluster) -> float:
        """Get the complete linkage error of merging this cluster with another.

        :return: max error between any two members of self | other
        """
        return float(np.max(self.members.pmatrix[np.ix_(self.ixs, other.ixs)]))

    # ===========================================================================
    #   the only verb
    # ===========================================================================

    def split(self) -> tuple[Cluster, Cluster]:
        """Split cluster into two clusters.

        :return: two new clusters
        :raises ValueError: if cluster has only one member

        Split the cluster into two clusters by the plane perpendicular to the axis of
        highest variance.

        The splitting is a bit funny due to innate characteristice of the stacked
        median. It is possible to get a split with members
            a) on one side of the splitting plane; and
            b) exactly on the splitting plane.
        See stacked_quantile module for details, but that case is covered here.
        """
        abc = self._direction_of_highest_variance
        vecs = self.members.vectors

        def rel_dist(x: int) -> float:
            return np.dot(abc, self.members.vectors[x])

        scored = sorted([(rel_dist(x), tuple(vecs[x]), x) for x in self.ixs])
        split = _split_floats([s for s, *_ in scored])
        if split in {0, len(scored)}:
            split = len(scored) // 2
        return (
            Cluster(self.members, [x for *_, x in scored[:split]]),
            Cluster(self.members, [x for *_, x in scored[split:]]),
        )



"""Paths to project binary files.

:author: Shay Hill
:created: 2022-10-25
"""

from pathlib import Path

_PROJECT_ROOT = Path(__file__).parent.parent.parent

BINARIES_DIR = _PROJECT_ROOT / "binaries"
TEST_DIR = _PROJECT_ROOT / "tests"

TEST_DIR.mkdir(parents=True, exist_ok=True)

BINARIES_DIR.mkdir(parents=True, exist_ok=True)



"""The members, clusters, and groups of clusters.

SuperclusterBase is an abstract class that can split or merge clusters.

SuperclusterBase caches states (lists of cluster indices, each defining a Cluster,
given a number of clusters) when splitting or merging.

A supercluster that starts as one large cluster will cache states as that cluster
and its descendants are split, and merging from any state in that cluster will be
loading a previouly cached state.

Similarly, a supercluster that starts as singletons will cache states as those
singletons and their descendants are merged, and splitting from any state in that
cluster will be loading a previously cached state.

The result of this is that a supercluster started as one large cluster will never
merge (only split and un-split) and a supercluster started as singletons will
never split (only merge and un-merge). The only thing required to make this a
divisive or agglomerative class is to implement the `_initialize_clusters` method
to return either a single cluster or a cluster for each member.

:author: Shay Hill
:created: 2023-01-17
"""

from __future__ import annotations

import itertools as it
from contextlib import suppress
from typing import TYPE_CHECKING, Literal, TypeVar

import numpy as np
from paragraphs import par

from cluster_colors.cluster_cluster import Cluster
from cluster_colors.cluster_members import Members
from cluster_colors.exceptions import FailedToMergeError, FailedToSplitError

_CachedState = tuple[tuple[int, ...], ...]

_RGB = tuple[float, float, float]


if TYPE_CHECKING:
    from collections.abc import Callable

    from cluster_colors.type_hints import (
        CenterName,
        CentroidName,
        FPArray,
        ProximityMatrix,
        QualityMetric,
        Vectors,
        VectorsLike,
    )


_SuperclusterT = TypeVar("_SuperclusterT", bound="SuperclusterBase")


class SuperclusterBase:
    """A list of Cluster instances.

    quality_metric: The property used to select which cluster to split. This is
        meaningles for agglomerative clustering. There is no such class attribute for
        merge quality, because I have only implemented one merge-quality metric:
        span.
        Literal["sum_error", "max_error", "avg_error", "span"]

    quality_centroid: The property used as a centroid when computing the quality
        metric.
        Literal["weighted_medoid", "unweighted_medoid"]

    assignment_centroid: The property used as a centroid when assigning members to a
        cluster. This may be different from the quality centroid, as SuperclusterBase
        deviates from k-medoids by splitting a cluster by median cut instead of by
        randomly selecting exemplars from the cluster.
        Literal["weighted_medoid", "unweighted_medoid"]

    clustering_method: The method used to split or merge clusters.
        Literal["divisive", "agglomerative"]
    """

    quality_metric: QualityMetric = "sum_error"
    quality_centroid: CentroidName = "weighted_medoid"
    assignment_centroid: CentroidName = "weighted_medoid"
    clustering_method: Literal["divisive", "agglomerative"] = "divisive"

    def __init__(self, members: Members) -> None:
        """Create a new Supercluster instance.

        :param members: Members instance
        """
        self.members = members

        class _Cluster(Cluster):
            quality_metric = self.quality_metric
            quality_centroid = self.quality_centroid

        self.cluster_type = _Cluster

        self.clusters: list[Cluster] = self._initialize_clusters()
        self._cached_states: list[_CachedState] = []
        self._cache_current_state()

    def _initialize_clusters(self) -> list[Cluster]:
        """Create clusters from the members."""
        match self.clustering_method:
            case "divisive":
                return [self.cluster_type(self.members, range(len(self.members)))]
            case "agglomerative":
                return [
                    self.cluster_type(self.members, [i])
                    for i in range(len(self.members))
                ]
        msg = par(
            f"""SuperclusterBase class attribute `clustering_type` must be 'divisive'
            or 'agglomerative', not {self.clustering_type}."""
        )
        raise ValueError(msg)

    # ===========================================================================
    #   constructors
    # ===========================================================================

    @classmethod
    def from_vectors(
        cls: type[_SuperclusterT],
        vectors: VectorsLike,
        pmatrix: ProximityMatrix | None = None,
    ) -> _SuperclusterT:
        """Create a SuperclusterBase instance from a sequence or array of colors.

        :param vectors: An iterable of vectors
            [(r0, g0, b0), (r1, g1, b1), ...]
        :param pmatrix: optional proximity matrix. If not given, will be calculated
            with squared Euclidean distance
        :return: A SuperclusterBase instance with members
            {Member([r0, g0, b0, 1]), Member([r1, g1, b1, 1]), ...}
        """
        members = Members.from_vectors(vectors, pmatrix=pmatrix)
        return cls(members)

    @classmethod
    def from_stacked_vectors(
        cls: type[_SuperclusterT],
        stacked_vectors: VectorsLike,
        pmatrix: ProximityMatrix | None = None,
    ) -> _SuperclusterT:
        """Create a Cluster instance from an array of colors with a weight axis.

        :param stacked_vectors: An iterable of vectors with a weight axis
            [(r0, g0, b0, w0), (r1, g1, b1, w1), ...]
        :param pmatrix: optional proximity matrix. If not given, will be calculated
            with squared Euclidean distance
            [(r0, g0, b0, w0), (r1, g1, b1, w1), ...]
        :return: A SuperclusterBase instance with members
            {Member([r0, g0, b0, w0]), Member([r1, g1, b1, w1]), ...}
        """
        members = Members.from_stacked_vectors(stacked_vectors, pmatrix=pmatrix)
        return cls(members)

    @classmethod
    def from_cluster_subset(cls, *clusters: Cluster) -> SuperclusterBase:
        """Create a SuperclusterBase instance from a subset of clusters.

        :param clusters: any number of clusters sharing the same members
        :return: SuperclusterBase instance with members from the subset

        Create a new SuperclusterBase instance from any number of clusters sharing
        the same Members instance. These clusters will presumably be a subset of the
        clusters from another SuperclusterBase instance.

        This filters the shared proximity matrix of the clusters to avoid calculating
        it again. The input clustering will be discarded, and the returned
        SuperclusterBase instance will in a 1-cluster or all-singletons state,
        depending on how _initialize_clusters is implemented.
        """
        shared_members = clusters[0].members
        if not all(c.members is shared_members for c in clusters[1:]):
            msg = "All clusters must share the same Members instance."
            raise ValueError(msg)
        ixs = sorted(int(x) for c in clusters for x in c.ixs)
        if len(ixs) != len(set(ixs)):
            msg = "Input clusters share member indices."
            raise ValueError(msg)
        subset_vectors = shared_members.vectors[ixs]
        subset_pmatrix = shared_members.pmatrix[np.ix_(ixs, ixs)]
        subset_members = Members(subset_vectors, pmatrix=subset_pmatrix)
        return cls(subset_members)

    # ===========================================================================
    #   properties
    # ===========================================================================

    @property
    def n(self) -> int:
        """Return the number of clusters in the Supercluster instance."""
        return len(self.clusters)

    def get_as_stacked_vectors(self, which_center: CenterName | None = None) -> Vectors:
        """Return the members as a numpy array, sorted heaviest to lightest.

        :param which_center: optionally specify a cluster center attribute. Choices
            are 'weighted_median', 'weighted_medoid', or 'unweighted_medoid'. Default
            is 'weighted_median'.
        :return as_stacked_vectors: members as a numpy array (n, m+1) with the last
            column as the weight.
        """
        as_stacked_vectors = np.array(
            [c.get_as_stacked_vector(which_center) for c in self.clusters]
        )
        return as_stacked_vectors[np.argsort(as_stacked_vectors[:, -1])][::-1]

    def get_as_vectors(self, which_center: CenterName | None = None) -> FPArray:
        """Return the members as a numpy array, sorted heaviest to lightest.

        :param which_center: optionally specify a cluster center attribute. Choices
            are 'weighted_median', 'weighted_medoid', or 'unweighted_medoid'. Default
            is 'weighted_median'.
        """
        return self.get_as_stacked_vectors(which_center)[:, :-1]

    # ===========================================================================
    #   cacheing and state management
    # ===========================================================================

    def _cache_current_state(self) -> None:
        """Cache the current state of the Supercluster instance.

        Call this at init and after every split or merge. These calls are already in
        the existing methods.
        """
        try:
            _ = self._get_cached_state(self.n)
        except IndexError:
            self._cached_states.append(tuple(tuple(c.ixs) for c in self.clusters))

    def _get_cached_state(self, n: int) -> _CachedState:
        """Get the cached state of the Supercluster with n clusters.

        :param n: number of clusters in the state
        :return: the state with n clusters
        :raise IndexError: if the state has not been cached

        This uses an indexing mechanic that will work with either divisive or
        agglomerative clustering.
        """
        idx = abs(n - len(self._cached_states[0]))
        try:
            return self._cached_states[idx]
        except IndexError as e:
            msg = f"State {n} has not been cached."
            raise IndexError(msg) from e

    def _restore_cached_state(self, state: _CachedState) -> None:
        """Restore a previous state of the Supercluster instance.

        :param state: state to restore
        :raise IndexError: if the state has not been cached

        Retains shared clusters between the current state and cached state to
        preserve cached values and relative values of cluster serial numbers.
        """
        current_state = tuple(tuple(c.ixs) for c in self.clusters)
        new_state = [x for x in state if x not in current_state]
        self.clusters = [c for c in self.clusters if tuple(c.ixs) in state]
        self.clusters.extend([self.cluster_type(self.members, x) for x in new_state])

    def _restore_state_to_n(self, n: int) -> None:
        """Restore the Supercluster instance to n clusters.

        :param n: desired number of clusters
        """
        if n == self.n:
            return
        state = self._get_cached_state(n)
        self._restore_cached_state(state)

    def _restore_state_as_close_as_possible_to_n(self, n: int) -> None:
        """Restore the Supercluster to the nearest state to n clusters.

        :param n: desired number of clusters

        If as state has not been cached with the desired number of clusters, get as
        close as possible.
        """
        with suppress(IndexError):
            self._restore_state_to_n(n)
            return
        state = self._cached_states[-1]
        if len(state) == self.n:
            return
        self._restore_cached_state(state)

    # ===========================================================================
    #   select clusters to split or merge
    # ===========================================================================

    def _get_next_to_split(self) -> Cluster:
        """Return the next set of clusters to split.

        :return: set of clusters with sse == max(sse)
        :raise ValueError: if no clusters are available to split

        Avoid picking singletons, which can tie for max_error == 0 with larger
        clusters if the larger clusters have 0-weight members.

        This function should never be called if all clusters are singletons, a
        FailedToSplitError should be raised before that happens.
        """
        candidates = (c for c in self.clusters if len(c.ixs) > 1)
        return max(candidates, key=lambda c: c.sum_error)

    def _get_next_to_merge(self) -> tuple[Cluster, Cluster]:
        """Return the next set of clusters to merge.

        :return: set of clusters with sse == min(sse)
        :raise ValueError: if no clusters are available to merge
        """
        if len(self.clusters) == 1:
            raise FailedToMergeError
        pairs = it.combinations(self.clusters, 2)
        return min(pairs, key=lambda p: p[0].get_merge_span(p[1]))

    # ===========================================================================
    #   perform splits and merges
    # ===========================================================================

    def _split_to_n(self, n: int) -> None:
        """Split or restore the Supercluster instance to n clusters.

        :param n: number of clusters
        """
        self._restore_state_as_close_as_possible_to_n(n)

        while self.n < n:
            if self.n == len(self.members):
                raise FailedToSplitError
            cluster = self._get_next_to_split()
            self.clusters.remove(cluster)
            self.clusters.extend(cluster.split())
            self._reassign()
            self._cache_current_state()

    def _merge_to_n(self, n: int) -> None:
        """Merge or restore the Supercluster instance to n clusters.

        :param n: number of clusters
        """
        self._restore_state_as_close_as_possible_to_n(n)
        while self.n > n:
            if self.n == 1:
                raise FailedToMergeError
            pair_to_merge = self._get_next_to_merge()
            merged_ixs = np.concatenate([x.ixs for x in pair_to_merge])
            merged = self.cluster_type(self.members, merged_ixs)
            self.clusters = [c for c in self.clusters if c not in pair_to_merge]
            self.clusters.append(merged)

    # ===========================================================================
    #   common public methods
    # ===========================================================================

    def set_n(self, n: int) -> None:
        """Set the number of clusters in the Supercluster instance.

        :param n: number of clusters
        """
        self._split_to_n(n)
        self._merge_to_n(n)

    def split(self):
        """Split the cluster with the highest sum error.

        This sets the state of the Supercluster instance. If the state is already
        >=n, nothing happens.
        """
        self._split_to_n(self.n + 1)

    def merge(self):
        """Merge the two clusters with the lowest sum error.

        This sets the state of the Supercluster instance. If the state is already
        <=n, nothing happens.
        """
        self._merge_to_n(self.n - 1)

    # ===========================================================================
    #   split or merge to satisfy a condition
    #
    #   For every condition defined here, the condition will be satisfied when a
    #   one-cluster state is reached (if not before), but other conditions may be
    #   passed in that to do satisfy this. In those instances, the splitting or
    #   merging will silenty give up when the minimum or maximum number of clusters
    #   is reached.
    # ===========================================================================

    def split_till_true(self, predicate: Callable[[SuperclusterBase], bool]) -> None:
        """Split clusters till predicate is True.

        :param predicate: function that takes a SuperclusterBase and returns a
            boolean. Presumably, if the predicate is True, a quality metric is good.
            If False, that metric is bad.

        If the predicate is True at the current state, merge till it is False, then
        split back to the first True. This will put the SuperclusterInstance into a
        state where the predicate is True, but one merge would make it False.

        This is for predicates that improve as the number of clusters increases.
        """
        with suppress(FailedToMergeError):
            while predicate(self):
                self.merge()
        with suppress(FailedToSplitError):
            while not predicate(self):
                self.split()

    def get_max_sum_error(self) -> float:
        """Return the maximum sum of errors of any cluster."""
        return max(c.span for c in self.clusters)

    def set_max_sum_error(self, max_sum_error: float):
        """Split as far as necessary to get below the threshold.

        :param min_proximity: maximum sum of errors of any cluster
        """

        def predicate(supercluster: SuperclusterBase) -> bool:
            return supercluster.get_max_sum_error() <= max_sum_error

        self.split_till_true(predicate)

    def get_max_span(self) -> float:
        """Return the minimum maximum cost of any cluster."""
        return max(c.span for c in self.clusters)

    def set_max_span(self, max_span: float):
        """Split as far as necessary to get below the threshold.

        :param min_max_error: maximum span of any cluster
        """

        def predicate(supercluster: SuperclusterBase) -> bool:
            return supercluster.get_max_span() <= max_span

        self.split_till_true(predicate)

    def get_max_max_error(self) -> float:
        """Return the maximum max_error of any cluster."""
        return max(c.max_error for c in self.clusters)

    def set_max_max_error(self, max_max_error: float):
        """Split as far as necessary to get below the threshold.

        :param min_max_error: maximum max_error of any cluster
        """

        def predicate(supercluster: SuperclusterBase) -> bool:
            return supercluster.get_max_max_error() <= max_max_error

        self.split_till_true(predicate)

    def get_max_avg_error(self) -> float:
        """Return the maximum avg_error of any cluster."""
        return max(c.avg_error for c in self.clusters)

    def set_max_avg_error(self, max_avg_error: float):
        """Split as far as necessary to get below the threshold.

        :param max_avg_error: maximum avg_error of any cluster
        """

        def predicate(supercluster: SuperclusterBase) -> bool:
            return supercluster.get_max_avg_error() <= max_avg_error

        self.split_till_true(predicate)

    # ===========================================================================
    #   the reassignment step for divisive clustering
    # ===========================================================================

    def _reassign(self, _previous_medoids: set[tuple[int, ...]] | None = None):
        """Reassign members based on proximity to cluster medoids.

        :param _previous_medoids: set of cluster medoids that have already been seen.
            For recursion use only

        Recursively redistribute members between clusters until no member can be
        moved to a different cluster to reduce the total error.

        A record of previous states prevents infinite recursion between a few states.
        It is conceivable that conversion could fail in other cases. The recursion
        limit is set to the Python's recursion limit.

        This will only ever be called for divisive clustering.
        """
        medoids = [c.unweighted_medoid for c in self.clusters]

        previous_states = _previous_medoids or set()
        state = tuple(sorted(medoids))
        if state in previous_states:
            return
        previous_states.add(state)

        which_medoid = np.argmin(self.members.pmatrix[medoids], axis=0)

        for i, cluster in enumerate(tuple(self.clusters)):
            new_where = np.argwhere(which_medoid == i)
            new = list(map(int, new_where.flatten()))
            if new != list(cluster.ixs):
                self.clusters.remove(cluster)
                self.clusters.append(self.cluster_type(self.members, new))

        with suppress(RecursionError):
            self._reassign(previous_states)


class DivisiveSupercluster(SuperclusterBase):
    """A SuperclusterBase that uses divisive clustering."""

    quality_metric: QualityMetric = "avg_error"
    quality_centroid: CentroidName = "weighted_medoid"
    assignment_centroid: CentroidName = "weighted_medoid"
    clustering_method: Literal["divisive", "agglomerative"] = "divisive"


class AgglomerativeSupercluster(SuperclusterBase):
    """A SuperclusterBase that uses agglomerative clustering."""

    quality_metric: QualityMetric = "avg_error"
    quality_centroid: CentroidName = "weighted_medoid"
    assignment_centroid: CentroidName = "weighted_medoid"
    clustering_method: Literal["divisive", "agglomerative"] = "divisive"



"""Test cut_colors.py

:author: Shay Hill
:created: 2022-10-22
"""

# pyright: reportPrivateUsage=false
# pyright: reportUnknownMemberType=false
# pyright: reportUnknownVariableType=false
# pyright: reportMissingParameterType=false
# pyright: reportUnknownParameterType=false

from typing import Annotated

import numpy as np
import numpy.typing as npt
import pytest

from cluster_colors import cut_colors, vector_stacker as sv

ColorsArray = Annotated[npt.NDArray[np.float64], (-1, 3)]


@pytest.fixture(
    scope="function",
    params=[np.random.randint(0, 255, (1000, 3), dtype=np.uint8) for _ in range(10)],
)
def colors(request) -> ColorsArray:
    return request.param


class TestCutColors:
    def test_cut_colors(self, colors: ColorsArray):
        """Call cut_colors with 100_000 random colors and pass result to stack_vectors."""
        colors = sv.stack_vectors(colors)
        _ = cut_colors.cut_colors(colors, 512)



"""Reduce colors by averaging colors with the same n-bit representation.

A reasonable and deterministic way to reduce 24-bit colors (8 bits per channel,
16_777_216 possible colors) to 1, 8, 64, 512, 4096, 32_768, 262_144, or 2_097_152
possible colors without Scipy.

Two-axis vectors (presumably <gray, weight>) should pass through unaffected because
they shouldn't have more than 256 unique values.

:author: Shay Hill
:created: 2022-09-19
"""

# pyright: reportUnknownMemberType=false
# pyright: reportUnknownArgumentType=false

from collections.abc import Callable
from itertools import chain
from typing import Annotated

import numpy as np
from basic_colormath import float_to_8bit_int
from paragraphs import par

from cluster_colors.type_hints import FPArray, NBits

_8BitCube = Annotated[FPArray, (256, 256, 256, ...)]
_FReduce = Callable[[FPArray, tuple[int, ...]], FPArray]

_EightBits = 8


def _floats_to_8bit_ints(*floats: float) -> tuple[int, ...]:
    """Convert floats from [0, 255] to 8-bit integers.

    :param floats: floats in the range [0.0, 255.0]
    :return: a tuple of 8-bit integers in the range [0, 255]
    """
    return tuple(float_to_8bit_int(f) for f in floats)


def _pool(matrix: FPArray, kernel_shape: tuple[int, ...], func: _FReduce) -> FPArray:
    """Pool a multi-dimensional array of numbers or arrays of numbers.

    :param matrix: array of numbers or arrays of numbers
    :param kernel_shape: shape of the kernel to pool with
    :param func: function to reduce the kernel. Must accept an array and axes as a
        tuple of floats.
    :return: pooled array
    :raises ValueError: if the matrix is not a multiple of the kernel_shape in each
        dimension

    Given an array (*dims) and a kernel shape (*kernel_dims), pool the array by func
    to each kernel_shape subarray.

    This expects the matrix to be a multiple of the kernel_shape in each dimension.
    For example, start with a 12x12 matrix. Pool this with a matix of 3x3. The result
    will be a 4x4 matrix. If you pool this with a 2x2 kernel, you'll get a 6x6
    matrix. If you pool this with a 4x4 kernel, you'll get a 3x3 matrix.

    The value of func will determine how the kernel is reduced. For example, if you
    use sum, the [0,0] value of a 4x4 matrix pooled to a 2x2 matrix will be the sum
    of the 16 values in the 4x4 matrix.
    """
    if not all(v % k == 0 for v, k in zip(matrix.shape, kernel_shape, strict=False)):
        msg = par(
            f"""matrix shape {matrix.shape} is not a multiple of kernel shape
            {kernel_shape}"""
        )
        raise ValueError(msg)
    matrix_shape = matrix.shape[: len(kernel_shape)]
    vector_shape = matrix.shape[len(kernel_shape) :]
    folded_dims = [(v // k, k) for v, k in zip(matrix_shape, kernel_shape, strict=True)]
    pools_shape = tuple(chain(*folded_dims))
    reshaped = matrix.reshape(pools_shape + vector_shape)
    return func(reshaped, tuple(x * 2 + 1 for x in range(len(matrix_shape))))


def _pool_8bit_cube(colors: FPArray, nbits: NBits) -> FPArray:
    """Sum values by n-bit representation of their indices.

    :param colors: array of colors, with shape (256, 256, 256, 4)
    :param nbits: number of bits per channel
    :return: array of reduced colors, with shape (n, n, n, 4) where n is 2**nbits
    """
    block = 2 ** (8 - nbits)
    kernel_shape: tuple[int, ...] = tuple(block for _ in colors.shape[:-1])
    return _pool(colors, kernel_shape, np.sum)


def _fill_colorspace(colors: FPArray) -> FPArray:
    """Fill values by n-bit representation of their indices.

    :param colors: array of colors, with shape (-1, 4)
    :return: a 256x256x256x4 array, where for every (..., w) in colors,
        result[...] = (...*w, w)
        e.g.,
        for every (r, g, b, w) in colors,
        result[r, g, b] = (r*w, g*w, b*w, w)

    Will fill any number of color channels, assuming all are 8-bit.
    """
    num_axes = colors.shape[-1]
    colorspace_shape = (256,) * (num_axes - 1) + (num_axes,)
    colorspace = np.zeros(colorspace_shape, dtype="float")
    for color in colors:
        vss = color[:-1] * color[-1]
        ixs = _floats_to_8bit_ints(*color[:-1])
        colorspace[ixs] += np.concatenate((vss, color[-1:]), axis=-1)
    return colorspace


def pool_colors(colors: FPArray, nbits: NBits = 6) -> FPArray:
    """Reduce 8-bit colors (each with a weight) to a maximum of (2**nbits)**3 colors.

    :param colors: array of colors, with shape (..., 4)
    :param nbits: number of bits per channel
    :return: array of reduced colors, with shape (..., 4)

    Create a (256,) * color_channels matrix. Place each color at its rgb coordinate
    in that matrix. Scale these colors by their weight so that later on, sum /
    combined weight will give an average color.

    Sum colors at adjacent coordinates then divine rgb by total weight.

    Will skip this entirely if you have fewer than 2**nbits-per-channel colors.
    There's no reason to discard information if you already have a workable amount of
    colors.
    """
    num_axes = colors.shape[-1]
    max_colors = (2**nbits) ** (num_axes - 1)
    if nbits >= _EightBits or len(colors) <= max_colors:
        nbits = 8

    colorspace = _fill_colorspace(colors)
    colorspace = _pool_8bit_cube(colorspace, nbits)
    colors = colorspace.reshape(-1, num_axes)
    colors = colors[colors[:, -1] > 0]
    colors[:, :-1] /= colors[:, -1:]
    return colors



"""Add and manipulate a vector weight axis.

This project is built around combining identical vectors (presumably colors) into
single instances with a weight axis (reflecting the combined weight of the combined
vectors) then treating those single combined vectors as multiples. For instance:

(1, 2, 3), (1, 2, 3), (4, 5, 6) would be stored as

(1, 2, 3, 2), (4, 5, 6, 1) but still treated as if it were

(1, 2, 3), (1, 2, 3), (4, 5, 6)

When working with pngs, there may be no need to add a weight channel, as the alpha
channel will serve the same function. Each pixel's alpha value will be interpreted as
the weight of that pixel.

The functions in this module return float arrays, not uint8 arrays. The reason's
being that float arrays will go out of range instead of wrapping around, which is
what we want (so we can identify and address it outside the module).

:author: Shay Hill
:created: 2022-10-18
"""

# pyright: reportUnknownMemberType=false
# pyright: reportUnknownArgumentType=false

from typing import Any

import numpy as np
from numpy import typing as npt

from cluster_colors.type_hints import FPArray, Vectors


def add_weight_axis(
    vectors: npt.NDArray[np.number[Any]], weight: float = 255.0
) -> FPArray:
    """Add a weight axis to a vector of vectors.

    :param vectors: A vector of vectors with shape (..., n).
    :param weight: The weight to add to each vector in the vector of vectors.
    :return: A vector of vectors with a weight axis. (..., n + 1)
    :raise ValueError: If the weight is not a positive number.

    The default weight is 255, which is the maximum value of a uint8. This will
    reflect full opacity, which makes sense when working with color vectors.

    If these vectors will only ever be used to represent multiple, full instances,
    then the weight could be any value, as long as it is consistent. 1 might be a
    more intuitive value in that case, as a vector with v[-1] == n would be a vector
    with n instances.
    """
    if weight <= 0:
        msg = f"Weight must be greater than 0. Got {weight}."
        raise ValueError(msg)
    ws = np.ones(vectors.shape[0]).reshape(-1, 1) * weight
    return np.hstack([vectors, ws]).astype(float)


def stack_vectors(
    vectors: npt.NDArray[np.number[Any]], weight: float | None = None
) -> Vectors:
    """Find and count unique vectors.

    :param vectors: array of numbers, with shape (..., n)
    :param weight: optionally provide a weight axis value.
        If not supplied, will assume last axis of each vector is a weight.
    :return: unique (by v[:-1]) with
        v[-1] equal to the sum of all v[-1] where v[:-1] == v[:-1]
    """
    if weight is not None:
        vectors = add_weight_axis(vectors, weight)

    flat_vectors = vectors.reshape(-1, vectors.shape[-1]).astype(float)

    unique_vectors, where_seen = np.unique(
        flat_vectors[:, :-1], return_inverse=True, axis=0
    )
    idx2seen = [0.0] * len(unique_vectors)
    for i, idx in enumerate(where_seen):
        idx2seen[idx] += flat_vectors[i, -1]
    weights = np.array(idx2seen).reshape(-1, 1)
    return np.append(unique_vectors, weights, axis=-1)
